import pandas as pd
from datetime import datetime, timedelta
import numpy as np


def filter_on_start_and_end_date(data, start_time, end_time):
    data['START_DATE'] = pd.to_datetime(data['START_DATE'], format='%Y/%m/%d %H:%M:%S')
    data['END_DATE'] = np.where(data['END_DATE'] == '2999/12/31 00:00:00', '2100/12/31 00:00:00', data['END_DATE'])
    data['END_DATE'] = pd.to_datetime(data['END_DATE'], format='%Y/%m/%d %H:%M:%S')
    data = data[(data['START_DATE'] < end_time) & (data['END_DATE'] > start_time)]
    return data


def filter_on_effective_date(data, start_time, end_time):
    data['EFFECTIVEDATE'] = pd.to_datetime(data['EFFECTIVEDATE'], format='%Y/%m/%d %H:%M:%S')
    data = data[data['EFFECTIVEDATE'] < end_time]
    return data


def filter_on_settlementdate(data, start_time, end_time):
    data['SETTLEMENTDATE'] = pd.to_datetime(data['SETTLEMENTDATE'], format='%Y/%m/%d %H:%M:%S')
    data = data[(data['SETTLEMENTDATE'] > start_time) & (data['SETTLEMENTDATE'] <= end_time)]
    return data


def filter_on_timestamp(data, start_time, end_time):
    data['TIMESTAMP'] = pd.to_datetime(data['TIMESTAMP'], format='%Y/%m/%d %H:%M:%S')
    data = data[(data['TIMESTAMP'] > start_time) & (data['TIMESTAMP'] <= end_time)]
    return data


def filter_on_interval_datetime(data, start_time, end_time):
    data['INTERVAL_DATETIME'] = pd.to_datetime(data['INTERVAL_DATETIME'], format='%Y/%m/%d %H:%M:%S')
    data = data[(data['INTERVAL_DATETIME'] > start_time) & (data['INTERVAL_DATETIME'] <= end_time)]
    return data


# Not tested, just for nemlite integration.
def filter_on_date_and_peroid(data, start_time, end_time):
    data = construct_interval_datetime_from_period_id(data)
    data = data[(data['SETTLEMENTDATE']  > start_time  ) & (data['SETTLEMENTDATE'] - timedelta(minutes=30) <= end_time)]
    return data


# Not tested, just for nemlite integration.
def filter_on_date_and_interval(data, start_time, end_time):
    data['SETTLEMENTDATE'] = pd.to_datetime(data['SETTLEMENTDATE'], format='%Y/%m/%d %H:%M:%S')
    data = data[(data['SETTLEMENTDATE'] > start_time) & (data['SETTLEMENTDATE'] <= end_time)]
    return data


# Not tested, just for nemlite integration.
def filter_on_last_changed(data, start_time, end_time):
    data['LASTCHANGED'] = pd.to_datetime(data['LASTCHANGED'], format='%Y/%m/%d %H:%M:%S')
    data = data[data['LASTCHANGED'] < end_time]
    return data


def filter_on_column_value(data, filter_cols, filter_values):
    for filter_col, filter_values in zip(filter_cols, filter_values):
        if filter_values is not None:
            data = data[data[filter_col].isin(filter_values)]
    return data

# Not tested, just for nemlite integration.
def construct_interval_datetime_from_period_id(data):
    data['SETTLEMENTDATE'] = np.vectorize(date_2_interval_datetime)(data['SETTLEMENTDATE'], data['PERIODID'])
    return data

# Not tested, just for nemlite integration.
def date_2_interval_datetime(date, period):
    datetime_obj = datetime.strptime(date, '%Y/%m/%d %H:%M:%S')
    datetime_obj = datetime(year=datetime_obj.year, month=datetime_obj.month, day=datetime_obj.day, hour=4)
    datetime_obj = datetime_obj + timedelta(minutes=(float(period)*30))
    return datetime_obj





import os
from nemosis import defaults

def write_file_names(name, month, year, day, index, raw_data_location):
    # Add the year and month information to the generic AEMO file name
    filename_full = defaults.names[name] + "_" + str(year) + str(month) + "010000.CSV"
    # Check if file already exists in the raw data folder.
    path_and_name = raw_data_location + '/' + filename_full
    # Add the year and month information to the generic AEMO file name
    filename_full_feather = defaults.names[name] + "_" + str(year) + str(month) + "010000.feather"
    # Check if file already exists in the raw data folder.
    path_and_name_feather = raw_data_location + '/' + filename_full_feather
    return filename_full, path_and_name, filename_full_feather, path_and_name_feather


def write_file_names_fcas(name, month, year, day, index, raw_data_location):
    # Add the year and month information to the generic AEMO file name
    filename_full = defaults.names[name] + "_" + str(year) + str(month) + day + index + ".CSV"
    # Check if file already exists in the raw data folder.
    path_and_name = raw_data_location + '/' + filename_full
    # Add the year and month information to the generic AEMO file name
    filename_full_feather = defaults.names[name] + "_" + str(year) + str(month) + day + index + ".feather"
    # Check if file already exists in the raw data folder.
    path_and_name_feather = raw_data_location + '/' + filename_full_feather
    return filename_full, path_and_name, filename_full_feather, path_and_name_feather


from nemosis import rows
from nemosis import defaults
from nemosis import data_fetch_methods
import pandas as pd
import tkinter as tk
import tkinter.ttk as ttk
from tkinter import messagebox
from tkinter import filedialog
import pickle
import traceback
import os
import sys


class VerticalScrollFrame(ttk.Frame):
    """A ttk frame allowing vertical scrolling only.
    Adapted from https://github.com/sunbearc22/tkinterWidgets/blob/master/scrframe.py.
    """

    def __init__(self, parent, *args, **options):
        mainborderwidth = options.pop('mainborderwidth', 0)
        interiorborderwidth = options.pop('interiorborderwidth', 0)
        mainrelief = options.pop('mainrelief', 'flat')
        interiorrelief = options.pop('interiorrelief', 'flat')
        ttk.Frame.__init__(self, parent, style='main.TFrame', borderwidth=mainborderwidth, relief=mainrelief)
        self.__createWidgets(interiorborderwidth, interiorrelief)
        self.canvas.bind('<Configure>', self.update_scrollbar)

    def __createWidgets(self, interiorborderwidth, interiorrelief):
        self.vscrollbar = ttk.Scrollbar(self, orient='vertical', style='canvas.Vertical.TScrollbar')
        self.vscrollbar.pack(side='right', fill='y', expand='false')
        self.canvas = tk.Canvas(self, yscrollcommand=self.vscrollbar.set, highlightthickness=0)
        self.canvas.pack(side='left', fill='both', expand='true')
        self.vscrollbar.config(command=self.canvas.yview)

        # reset the view
        self.canvas.xview_moveto(0)
        self.canvas.yview_moveto(0)

        # create a frame inside the canvas which will be scrolled with it
        self.interior = ttk.Frame(self.canvas, borderwidth=interiorborderwidth, relief=interiorrelief)
        self.interior_id = self.canvas.create_window(0, 0, window=self.interior, anchor='nw',)

    def update_scrollbar(self, event):
        '''Configure the interior frame size and the canvas scrollregion'''
        # Force the update of .winfo_width() and winfo_height()
        self.canvas.update_idletasks()

        # Internal parameters
        interiorReqHeight = self.interior.winfo_reqheight()
        canvasWidth = self.canvas.winfo_width()
        canvasHeight = self.canvas.winfo_height()

        # Set interior frame width to canvas current width
        self.canvas.itemconfigure(self.interior_id, width=canvasWidth)

        # Set interior frame height and canvas scrollregion
        if canvasHeight > interiorReqHeight:
            self.canvas.itemconfigure(self.interior_id, height=canvasHeight)
            self.canvas.config(scrollregion="0 0 {0} {1}".
                               format(canvasWidth, canvasHeight))
        else:
            self.canvas.itemconfigure(self.interior_id, height=interiorReqHeight)
            self.canvas.config(scrollregion="0 0 {0} {1}".
                               format(canvasWidth, interiorReqHeight))


class App(ttk.Frame):

    def __init__(self, parent, *args, **kwargs):

        ttk.Frame.__init__(self, parent=None, style='App.TFrame', borderwidth=0, width=890, height=590)
        self.parent = parent
        self.parent.title('NEMOSIS')
        self.parent.geometry('1000x600')
        self.setStyle()
        self.createWidgets()
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

    def setStyle(self):
        style = ttk.Style()
        style.configure('App.TFrame', background='pink')

    def createWidgets(self):
        self.frame = VerticalScrollFrame(self, arrowcolor='white',  mainborderwidth=10, interiorborderwidth=10,
                                         mainrelief='raised', interiorrelief='sunken')

        self.frame.grid(row=0, column=0, sticky='nsew')
        self.rows = []
        self.add_header()
        self.add_plus()
        self.add_AEMO_query()

    def add_header(self):
        # Create the default starting widgets that appear at the top of the gui.
        self.header = ttk.Frame(self.frame.interior)
        self.header.grid(row=0, column=0, columnspan=50, sticky='w')
        self.header.update()

        # Label for save location entry box.
        self.save_label = tk.Label(self.header, text='  Output data to:', anchor='w')
        self.save_label.grid(row=0, column=1)
        self.save_label.config(width=15)
        self.save_label.update()

        # Text entry that specifies the location to save query results.
        self.save_location = ttk.Entry(self.header)
        self.save_location.grid(row=0, column=2, columnspan=defaults.save_field_column_span)
        self.save_location.config(width=50)
        self.save_location.update()

        # Button set save location.
        self.output_location = ttk.Button(self.header, text='...', command=self.set_save_location)
        self.output_location.grid(row=0, column=5)
        self.output_location.config(width=4)
        self.output_location.update()

        # Label for the raw data location entry box.
        self.raw_data_label = ttk.Label(self.header, text='Raw data cache:', anchor='w')
        self.raw_data_label.grid(row=1, column=1)
        self.raw_data_label.config(width=15)
        self.raw_data_label.update()

        # Text entry that specifies the location of the raw aemo data cache.
        self.raw_data_location = ttk.Entry(self.header)
        self.raw_data_location.grid(row=1, column=2, columnspan=defaults.save_field_column_span)
        self.raw_data_location.config(width=50)
        self.raw_data_location.update()

        # Button set save location.
        self.output_location = ttk.Button(self.header, text='...', command=self.set_cache_location)
        self.output_location.grid(row=1, column=5)
        self.output_location.config(width=4)
        self.output_location.update()

        # Button to save current state of the gui.
        self.save = ttk.Button(self.header, text='Save session', command=self.save_session)
        self.save.grid(row=0, column=6, padx=20)
        self.save.config(width=20)
        self.save.update()

        # Button to load a previous state of the gui.
        self.load = ttk.Button(self.header, text='Load session', command=self.load_session)
        self.load.grid(row=1, column=6, padx=20)
        self.load.config(width=20)
        self.load.update()

    def add_plus(self):
        # Add the button that added extra query and merge rows to the gui.
        self.row_adder = ttk.Frame(self.frame.interior)
        self.row_adder.grid(row=defaults.query_row_offset + len(self.rows) * defaults.row_height
                            + defaults.plus_internal_row, column=0, padx=defaults.standard_x_pad, sticky='w',
                            columnspan=50, pady=10)
        self.row_adder.update()

        # Button to add extra queries.
        self.plus_AEMO_query = ttk.Button(self.row_adder, text=u"\u2795" + ' AEMO table',
                                          command=self.add_AEMO_query)
        self.plus_AEMO_query.grid(row=0, column=0)
        self.plus_AEMO_query.update()

        # Button to add extra queries.
        self.plus_custom_query = ttk.Button(self.row_adder, text=u"\u2795" + ' Custom table',
                                            command=self.add_Custom_query)
        self.plus_custom_query.grid(row=0, column=1)
        self.plus_custom_query.update()

        # Button to add extra merge.
        self.plus_merge = ttk.Button(self.row_adder, text=u"\u2795" + ' Merge', command=self.add_merge)
        self.plus_merge.grid(row=0, column=3)
        self.plus_merge.update()

        # Button to run the app.
        self.run = ttk.Button(self.row_adder, text=u"\u25B6" + '  Run queries ', command=self.run_queries)
        self.run.grid(row=0, column=4)
        self.run.update()

    def add_AEMO_query(self):
        # Function to add extra query.
        self.rows.append(rows.Query(self.frame.interior, len(self.rows), self, table_options=defaults.display_as_AMEO))
        self.replace_plus()

    def add_Custom_query(self):
        # Function to add extra query.
        self.rows.append(rows.Query(self.frame.interior, len(self.rows), self, table_options=defaults.display_as_Custom))
        self.replace_plus()

    def add_merge(self):
        # Function to add extra merge.
        self.rows.append(rows.Merge(self.frame.interior, len(self.rows), self))
        self.replace_plus()

    def delete_row(self, row_number):
        # Function to delete row.
        # Remove widgets and delete row.
        self.rows[row_number].empty()
        del self.rows[row_number]
        # Reposition remaining rows of queries and merges so no gaps are left.
        for row_number, row in enumerate(self.rows):
            row.row_number = row_number
            row.position()
        self.frame.update_scrollbar(None)

    def run_queries(self):
        # Run all the rows in the gui.
        results = {}
        save_location = self.save_location.get()
        raw_data_location = self.raw_data_location.get()
        try:
            for row in self.rows:
                save_name = row.name.get()
                if type(row).__name__ == 'Query':
                    results[save_name] = self.run_query(row, raw_data_location)
                elif type(row).__name__ == 'Merge':
                    results[save_name] = self.run_merge(row, results)
                results[save_name].to_csv(save_location + '\\' + save_name + '.csv', index=False,
                                          date_format='%Y/%m/%d %H:%M:%S')
            messagebox.showinfo('Finished', 'Your query has finished!')
        except Exception:
                traceback.print_exc()
                messagebox.showerror('Error', 'Your query executed with an error. '
                                 '\nReview the console for detailed information')

        return

    def run_query(self, row, raw_data_location):
        # Run an individual query.

        # Find the table name from the row.
        table = row.table_options[row.tables.curselection()[0]]
        # Find the select columns.
        columns = tuple([row.col_list.get(0, tk.END)[index] for index in row.col_list.curselection()])
        # Find the columns that could be filtered on.
        potential_filter_cols = list(row.filter_label.keys())
        filter_cols = ()
        filter_values = ()
        # Build a list of filter columns and filter values if the filters list have any values in them.
        for column in potential_filter_cols:
            selection = [row.filter_list[column].get(0, tk.END)[index] for index in
                         row.filter_list[column].curselection()]
            if len(selection) > 0:
                    filter_values = filter_values + (selection,)
                    filter_cols = filter_cols + (column,)

        start_time = row.start_time.get()
        end_time = row.end_time.get()
        # Call the query using the tables predefined wraper function.
        result = data_fetch_methods.method_map[table](start_time, end_time, table, raw_data_location, columns,
                                                      filter_cols, filter_values)
        return result

    def run_merge(self, row, results):
        # Run an individual merge row.
        # Get the name of the result to put on the left of the merge.
        left_table_name = row.left_table.get()
        # Get the result to put on the left of the merge.
        left_table = results[left_table_name]
        # Get the keys to use on the right result.
        left_keys = [row.left_key_list.get(0, tk.END)[index] for index in row.left_key_list.curselection()]
        # Get the name of the result to put on the left of the merge.
        right_table_name = row.right_table.get()
        # Get the result to put on the right of the merge.
        right_table = results[right_table_name]
        # Get the keys to use on the right result.
        right_keys = [row.right_key_list.get(0, tk.END)[index] for index in row.right_key_list.curselection()]
        # Get the join type to use.
        join_type = defaults.join_type[row.join_types.curselection()[0]]
        # Merge the results.
        result = pd.merge(left_table, right_table, join_type, left_on=left_keys, right_on=right_keys)
        return result

    def set_save_location(self):
        save_name = filedialog.askdirectory()
        self.save_location.delete(0, "end")
        self.save_location.insert("end", save_name)

    def set_cache_location(self):
        save_name = filedialog.askdirectory()
        self.raw_data_location.delete(0, "end")
        self.raw_data_location.insert("end", save_name)

    def save_session(self):
        # Save the current state of the gui and pickle the result.
        session_state = {}
        session_state['raw_data_location'] = self.raw_data_location.get()
        session_state['save_location'] = self.save_location.get()
        session_state['rows'] = []
        for row in self.rows:
            session_state['rows'].append(row.state())
        save_name = filedialog.asksaveasfilename()

        # If a user provides a save name with .pkl already in it do not add another .pkl
        if save_name[-4:] != '.pkl':
            save_name = save_name + '.pkl'

        with open(save_name, 'wb') as f:
            pickle.dump(session_state, f, pickle.HIGHEST_PROTOCOL)

    def load_session(self, session=None):
        # Reconfigure the gui to match a state defined in a dictionary.

        # If the function is called with a session argument then use that (this is used for testing). If it is called
        # without a session argument then a popup asks the user to provide one.
        if session is None:
            save_name = filedialog.askopenfilename()
            with open(save_name, 'rb') as f:
                session_state = pickle.load(f)
        else:
            session_state = session

        # Load in raw data and save locations.
        self.raw_data_location.delete(0, 'end')
        self.raw_data_location.insert(0, session_state['raw_data_location'])
        self.save_location.delete(0, 'end')
        self.save_location.insert(0, session_state['save_location'])

        # Empty any existing rows.
        for row in self.rows:
            row.empty()
        # Create and empty row list to delete any existing rows.
        self.rows = []

        # Create rows based on the saved session.
        for row in session_state['rows']:
            # Create the right type of row.
            if row['type'] == 'query':
                if len(row['table']) != 0:
                    if row['table'] in defaults.display_as_AMEO:
                        self.rows.append(rows.Query(self.frame.interior, len(self.rows),
                                                    self, defaults.display_as_AMEO))
                    if row['table'] in defaults.display_as_Custom:
                        self.rows.append(rows.Query(self.frame.interior, len(self.rows),
                                                    self, defaults.display_as_Custom))

            elif row['type'] == 'merge':
                self.rows.append(rows.Merge(self.frame.interior, len(self.rows), self))
            # Load the row state.
            self.rows[-1].load_state(row)

        # Put the plus button at the buttom of all the rows.
        self.replace_plus()

    def replace_plus(self):
        # Move the plus buttons to below all existing rows.
        self.row_adder.grid(row=defaults.query_row_offset + (len(self.rows)) * defaults.row_height
                            + defaults.plus_internal_row)
        self.row_adder.update()
        self.frame.update_scrollbar(None)

def resource_path(relative_path):
    """ Get absolute path to resource, works for dev and for PyInstaller """
    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        base_path = sys._MEIPASS
    except Exception:
        base_path = os.path.abspath(".")

    return os.path.join(base_path, relative_path)


if __name__ == '__main__':
    import tempfile

    ICON = (b'\x00\x00\x01\x00\x01\x00\x10\x10\x00\x00\x01\x00\x08\x00h\x05\x00\x00'
            b'\x16\x00\x00\x00(\x00\x00\x00\x10\x00\x00\x00 \x00\x00\x00\x01\x00'
            b'\x08\x00\x00\x00\x00\x00@\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x01\x00\x00\x00\x01') + b'\x00' * 1282 + b'\xff' * 64

    _, ICON_PATH = tempfile.mkstemp()
    with open(ICON_PATH, 'wb') as icon_file:
        icon_file.write(ICON)


    root = tk.Tk()
    app = App(root)
    app.grid(row=0, column=0, sticky='nsew')
    root.rowconfigure(0, weight=1)
    root.columnconfigure(0, weight=1)
    root.iconbitmap(resource_path('favicon.ico'))
    root.mainloop()


import unittest
import pandas as pd
from nemosis import query_wrapers
from pandas.util.testing import assert_frame_equal
from datetime import datetime
from nemosis import defaults


class TestDispatchDateSetup(unittest.TestCase):
    def setUp(self):
        self.start_time = '2017/01/01 00:10:00'
        self.end_time = '2019/06/03 12:15:00'

    def test_start_time_and_end_time(self):
        start_time, end_time = query_wrapers.dispatch_date_setup(self.start_time, self.end_time)
        self.assertEqual(start_time, '2016/12/30 23:59:59')
        self.assertEqual(end_time, '2019/06/03 00:00:00')


class TestFCASFinalise(unittest.TestCase):
    def setUp(self):
        self.testDataFrame1 = pd.DataFrame({'StingsWithSpaces': [' HELLO ', 'foo and baa ', ' at front']})
        self.testDataFrame2 = pd.DataFrame({'StingsWithSpaces': [' HELLO ', 'foo and baa ', ' at front'],
                                            'INTs2ignore': [1, 2, 3]})
        self.testDataFrame3 = pd.DataFrame({'StingsWithSpaces': [' HELLO ', 'foo and baa ', ' at front'],
                                            'INTs2ignore': [1, 2, 3],
                                            'StingsWithSpaces2': [' HELLO ', 'foo and baa ', ' at front']})

    def test_string_cleanup_one_col(self):
        result = query_wrapers.fcas4s_finalise(self.testDataFrame1, None, None)
        aim = pd.DataFrame({'StingsWithSpaces': ['HELLO', 'foo and baa', 'at front']})
        assert_frame_equal(aim, result)

    def test_string_cleanup_one_col_ignore_one_col(self):
        result = query_wrapers.fcas4s_finalise(self.testDataFrame2, None, None)
        aim = pd.DataFrame({'StingsWithSpaces': ['HELLO', 'foo and baa', 'at front'],
                            'INTs2ignore': [1, 2, 3]})
        assert_frame_equal(aim, result)

    def test_cleanup_ignore_cleanup(self):
        result = query_wrapers.fcas4s_finalise(self.testDataFrame3, None, None)
        aim = pd.DataFrame({'StingsWithSpaces': ['HELLO', 'foo and baa', 'at front'],
                            'INTs2ignore': [1, 2, 3],
                            'StingsWithSpaces2': ['HELLO', 'foo and baa', 'at front']})
        assert_frame_equal(aim, result)


class TestMostRecent(unittest.TestCase):
    def setUp(self):
        self.dummyGenConData = \
        pd.DataFrame({'EFFECTIVEDATE': ['2017/01/01 00:00:00', '2017/01/04 00:15:00', '2018/05/01 00:00:00'],
                      'VERSIONNO': ['5', '1', '1'],
                      'GENCONID': ['ID1', 'ID1', 'ID1']})
        self.dummyGenConData['EFFECTIVEDATE'] = pd.to_datetime(self.dummyGenConData['EFFECTIVEDATE'],
                                                               format='%Y/%m/%d %H:%M:%S')

        self.dummyGenConData2 = \
        pd.DataFrame({'EFFECTIVEDATE': ['2017/01/01 00:00:00', '2017/01/04 00:15:00', '2017/01/04 00:15:00'],
                      'VERSIONNO': ['5', '1', '2'],
                      'GENCONID': ['ID1', 'ID1', 'ID1']})
        self.dummyGenConData2['EFFECTIVEDATE'] = pd.to_datetime(self.dummyGenConData2['EFFECTIVEDATE'],
                                                                format='%Y/%m/%d %H:%M:%S')

        self.dummyGenConData3 = \
        pd.DataFrame({'EFFECTIVEDATE': ['2017/01/01 00:00:00', '2017/01/04 00:15:00', '2017/01/04 00:15:00',
                                        '2017/01/01 00:00:00', '2017/01/04 00:15:00', '2017/01/04 00:15:00'],
                      'VERSIONNO': ['5', '1', '2', '5', '1', '2'],
                      'GENCONID': ['ID1', 'ID1', 'ID1', 'ID2', 'ID2', 'ID2']})
        self.dummyGenConData3['EFFECTIVEDATE'] = pd.to_datetime(self.dummyGenConData3['EFFECTIVEDATE'],
                                                                format='%Y/%m/%d %H:%M:%S')

    def test_one_id_gencondata_start_date_after_all(self):
        start_time = datetime.strptime('2019/06/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        table_name = 'dummy'
        defaults.primary_date_columns['dummy'] = 'EFFECTIVEDATE'
        defaults.effective_date_group_col['dummy'] = ['GENCONID']
        result = query_wrapers.most_recent_records_before_start_time(self.dummyGenConData, start_time, table_name
                                                                     ).reset_index(drop=True)
        aim = pd.DataFrame({'EFFECTIVEDATE': ['2018/05/01 00:00:00'], 'VERSIONNO': ['1'], 'GENCONID': ['ID1']})
        aim['EFFECTIVEDATE'] = pd.to_datetime(aim['EFFECTIVEDATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_one_id_gencondata_start_date_after_first(self):
        start_time = datetime.strptime('2017/01/01 01:00:00', '%Y/%m/%d %H:%M:%S')
        table_name = 'dummy'
        defaults.primary_date_columns['dummy'] = 'EFFECTIVEDATE'
        defaults.effective_date_group_col['dummy'] = ['GENCONID']
        result = query_wrapers.most_recent_records_before_start_time(self.dummyGenConData, start_time, table_name).\
            sort_values('EFFECTIVEDATE', ascending=False)\
            .reset_index(drop=True)
        aim = pd.DataFrame({'EFFECTIVEDATE': ['2017/01/01 00:00:00', '2017/01/04 00:15:00', '2018/05/01 00:00:00'],
                            'VERSIONNO': ['5', '1', '1'], 'GENCONID': ['ID1', 'ID1', 'ID1']}
                           ).sort_values('EFFECTIVEDATE', ascending=False)\
            .reset_index(drop=True)
        aim['EFFECTIVEDATE'] = pd.to_datetime(aim['EFFECTIVEDATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_one_id_gencondata_start_date_before_all(self):
        start_time = datetime.strptime('2010/01/01 01:00:00', '%Y/%m/%d %H:%M:%S')
        table_name = 'dummy'
        defaults.primary_date_columns['dummy'] = 'EFFECTIVEDATE'
        defaults.effective_date_group_col['dummy'] = ['GENCONID']
        result = query_wrapers.most_recent_records_before_start_time(self.dummyGenConData, start_time, table_name).\
            sort_values('EFFECTIVEDATE', ascending=False).reset_index(drop=True)
        aim = pd.DataFrame({'EFFECTIVEDATE': ['2017/01/01 00:00:00', '2017/01/04 00:15:00', '2018/05/01 00:00:00'],
                            'VERSIONNO': ['5', '1', '1'],
                            'GENCONID': ['ID1', 'ID1', 'ID1']}).sort_values('EFFECTIVEDATE', ascending=False)\
            .reset_index(drop=True)
        aim['EFFECTIVEDATE'] = pd.to_datetime(aim['EFFECTIVEDATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_one_id_gencondata_and_repeated_effectivedate_start_date_after_all(self):
        start_time = datetime.strptime('2019/06/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        table_name = 'dummy'
        defaults.primary_date_columns['dummy'] = 'EFFECTIVEDATE'
        defaults.effective_date_group_col['dummy'] = ['GENCONID']
        result = query_wrapers.most_recent_records_before_start_time(self.dummyGenConData2, start_time, table_name
                                                                     ).reset_index(drop=True)
        aim = pd.DataFrame({'EFFECTIVEDATE': ['2017/01/04 00:15:00', '2017/01/04 00:15:00'], 'VERSIONNO': ['1', '2'],
                            'GENCONID': ['ID1', 'ID1']})
        aim['EFFECTIVEDATE'] = pd.to_datetime(aim['EFFECTIVEDATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)

    def test_2_id_and_repeated_effectivedate_start_date_after_all(self):
        start_time = datetime.strptime('2019/06/01 00:00:00', '%Y/%m/%d %H:%M:%S')
        table_name = 'dummy'
        defaults.primary_date_columns['dummy'] = 'EFFECTIVEDATE'
        defaults.effective_date_group_col['dummy'] = ['GENCONID']
        result = query_wrapers.most_recent_records_before_start_time(self.dummyGenConData3, start_time, table_name
                                                                     ).reset_index(drop=True)
        aim = pd.DataFrame({'EFFECTIVEDATE': ['2017/01/04 00:15:00', '2017/01/04 00:15:00',
                                              '2017/01/04 00:15:00', '2017/01/04 00:15:00'],
                            'VERSIONNO': ['1', '2', '1', '2'],
                            'GENCONID': ['ID1', 'ID1', 'ID2', 'ID2']})
        aim['EFFECTIVEDATE'] = pd.to_datetime(aim['EFFECTIVEDATE'], format='%Y/%m/%d %H:%M:%S')
        assert_frame_equal(aim, result)



import os

from nemosis import filters, downloader, query_wrapers, write_file_names, date_generators


setup = {'DISPATCHLOAD': None,
         'TRADINGLOAD': None,
         'TRADINGPRICE': None,
         'TRADINGREGIONSUM': None,
         'TRADINGINTERCONNECT': None,
         'DISPATCHPRICE': None,
         'DISPATCH_UNIT_SCADA': None,
         'DISPATCHCONSTRAINT': None,
         'DUDETAILSUMMARY': None,
         'DUDETAIL': None,
         'GENCONDATA': None,
         'SPDREGIONCONSTRAINT': None,
         'SPDCONNECTIONPOINTCONSTRAINT': None,
         'SPDINTERCONNECTORCONSTRAINT': None,
         'FCAS_4_SECOND': None,
         'ELEMENTS_FCAS_4_SECOND': None,
         'VARIABLES_FCAS_4_SECOND': None,
         'Generators and Scheduled Loads': None,
         'BIDDAYOFFER_D': query_wrapers.dispatch_date_setup,
         'BIDPEROFFER_D': None,
         'FCAS_4s_SCADA_MAP': None,
         'DISPATCHINTERCONNECTORRES': None,
         'DISPATCHREGIONSUM': None,
         'LOSSMODEL': None,
         'LOSSFACTORMODEL': None,
         'MNSP_DAYOFFER': query_wrapers.dispatch_date_setup,
         'MNSP_PEROFFER': query_wrapers.dispatch_half_hour_setup,
         'MNSP_INTERCONNECTOR': None,
         'INTERCONNECTOR': None,
         'INTERCONNECTORCONSTRAINT': None,
         'MARKET_PRICE_THRESHOLDS': None}

search_type = {'DISPATCHLOAD': 'start_to_end',
               'TRADINGLOAD': 'start_to_end',
               'TRADINGPRICE': 'start_to_end',
               'TRADINGREGIONSUM': 'start_to_end',
               'TRADINGINTERCONNECT': 'start_to_end',
               'DISPATCHPRICE': 'start_to_end',
               'DISPATCH_UNIT_SCADA': 'start_to_end',
               'DISPATCHCONSTRAINT': 'start_to_end',
               'DUDETAILSUMMARY': 'end',
               'DUDETAIL': 'all',
               'GENCONDATA': 'all',
               'SPDREGIONCONSTRAINT': 'all',
               'SPDCONNECTIONPOINTCONSTRAINT': 'all',
               'SPDINTERCONNECTORCONSTRAINT': 'all',
               'FCAS_4_SECOND': 'start_to_end',
               'ELEMENTS_FCAS_4_SECOND': None,
               'VARIABLES_FCAS_4_SECOND': None,
               'Generators and Scheduled Loads': None,
               'BIDDAYOFFER_D': 'start_to_end',
               'BIDPEROFFER_D': 'start_to_end',
               'FCAS_4s_SCADA_MAP': None,
               'DISPATCHINTERCONNECTORRES': 'start_to_end',
               'DISPATCHREGIONSUM': 'start_to_end',
               'LOSSMODEL': 'all',
               'LOSSFACTORMODEL': 'all',
               'MNSP_DAYOFFER': 'start_to_end',
               'MNSP_PEROFFER': 'start_to_end',
               'MNSP_INTERCONNECTOR': 'all',
               'INTERCONNECTOR': 'all',
               'INTERCONNECTORCONSTRAINT': 'all',
               'MARKET_PRICE_THRESHOLDS': 'all'}

filter = {'DISPATCHLOAD': filters.filter_on_settlementdate,
          'TRADINGLOAD': filters.filter_on_settlementdate,
          'TRADINGPRICE': filters.filter_on_settlementdate,
          'TRADINGREGIONSUM': filters.filter_on_settlementdate,
          'TRADINGINTERCONNECT': filters.filter_on_settlementdate,
          'DISPATCHPRICE': filters.filter_on_settlementdate,
          'DISPATCH_UNIT_SCADA': filters.filter_on_settlementdate,
          'DISPATCHCONSTRAINT': filters.filter_on_settlementdate,
          'DUDETAILSUMMARY': filters.filter_on_start_and_end_date,
          'DUDETAIL': filters.filter_on_effective_date,
          'GENCONDATA': filters.filter_on_effective_date,
          'SPDREGIONCONSTRAINT': filters.filter_on_effective_date,
          'SPDCONNECTIONPOINTCONSTRAINT': filters.filter_on_effective_date,
          'SPDINTERCONNECTORCONSTRAINT': filters.filter_on_effective_date,
          'FCAS_4_SECOND': filters.filter_on_timestamp,
          'ELEMENTS_FCAS_4_SECOND': None,
          'VARIABLES_FCAS_4_SECOND': None,
          'Generators and Scheduled Loads': None,
          'BIDDAYOFFER_D': filters.filter_on_settlementdate,
          'BIDPEROFFER_D': filters.filter_on_interval_datetime,
          'FCAS_4s_SCADA_MAP': None,
          'DISPATCHINTERCONNECTORRES': filters.filter_on_settlementdate,
          'DISPATCHREGIONSUM': filters.filter_on_settlementdate,
          'LOSSMODEL': filters.filter_on_effective_date,
          'LOSSFACTORMODEL': filters.filter_on_effective_date,
          'MNSP_DAYOFFER': filters.filter_on_settlementdate,
          'MNSP_PEROFFER': filters.filter_on_date_and_peroid,
          'MNSP_INTERCONNECTOR': filters.filter_on_effective_date,
          'INTERCONNECTOR': filters.filter_on_last_changed,
          'INTERCONNECTORCONSTRAINT': filters.filter_on_effective_date,
          'MARKET_PRICE_THRESHOLDS': filters.filter_on_effective_date}

finalise = {'DISPATCHLOAD': None,
            'TRADINGLOAD': None,
            'TRADINGPRICE': None,
            'TRADINGREGIONSUM': None,
            'TRADINGINTERCONNECT': None,
            'DISPATCHPRICE': None,
            'DISPATCH_UNIT_SCADA': None,
            'DISPATCHCONSTRAINT': [query_wrapers.convert_genconid_effectivedate_to_datetime_format],
            'DUDETAILSUMMARY': [query_wrapers.most_recent_records_before_start_time,
                                query_wrapers.drop_duplicates_by_primary_key],
            'DUDETAIL': [query_wrapers.most_recent_records_before_start_time,
                                query_wrapers.drop_duplicates_by_primary_key],
            'GENCONDATA': [query_wrapers.most_recent_records_before_start_time,
                           query_wrapers.drop_duplicates_by_primary_key],
            'SPDREGIONCONSTRAINT': [query_wrapers.most_recent_records_before_start_time],
            'SPDCONNECTIONPOINTCONSTRAINT': [query_wrapers.most_recent_records_before_start_time,
                                             query_wrapers.drop_duplicates_by_primary_key],
            'SPDINTERCONNECTORCONSTRAINT': [query_wrapers.most_recent_records_before_start_time,
                                            query_wrapers.drop_duplicates_by_primary_key],
            'FCAS_4_SECOND': [query_wrapers.fcas4s_finalise],
            'ELEMENTS_FCAS_4_SECOND': None,
            'VARIABLES_FCAS_4_SECOND': None,
            'Generators and Scheduled Loads': None,
            'BIDDAYOFFER_D': None,
            'BIDPEROFFER_D': None,
            'FCAS_4s_SCADA_MAP': None,
            'DISPATCHINTERCONNECTORRES': None,
            'DISPATCHREGIONSUM': None,
            'LOSSMODEL': [query_wrapers.most_recent_records_before_start_time,
                              query_wrapers.drop_duplicates_by_primary_key],
            'LOSSFACTORMODEL': [query_wrapers.most_recent_records_before_start_time,
                              query_wrapers.drop_duplicates_by_primary_key],
            'MNSP_DAYOFFER': [query_wrapers.most_recent_records_before_start_time,
                              query_wrapers.drop_duplicates_by_primary_key],
            'MNSP_PEROFFER': [query_wrapers.most_recent_records_before_start_time,
                              query_wrapers.drop_duplicates_by_primary_key],
            'MNSP_INTERCONNECTOR': [query_wrapers.most_recent_records_before_start_time,
                                    query_wrapers.drop_duplicates_by_primary_key],
            'INTERCONNECTOR': [query_wrapers.most_recent_records_before_start_time,
                               query_wrapers.drop_duplicates_by_primary_key],
            'INTERCONNECTORCONSTRAINT': [query_wrapers.most_recent_records_before_start_time,
                                         query_wrapers.drop_duplicates_by_primary_key],
            'MARKET_PRICE_THRESHOLDS': None}

date_gen = {'MMS': date_generators.year_and_month_gen,
            'FCAS': date_generators.year_month_day_index_gen}

write_filename = {'MMS': write_file_names.write_file_names,
                  'FCAS': write_file_names.write_file_names_fcas}

downloader = {'MMS': downloader.run,
              'FCAS': downloader.run_fcas4s}



from nemosis import defaults
from calendar import monthrange


def year_and_month_gen(start_time, end_time):

    if start_time.day == 1 and start_time.hour == 0 and start_time.minute == 0:
        if start_time.month == 1:
            start_time = start_time.replace(month=12)
            start_time = start_time.replace(year=start_time.year - 1)
        else:
            start_time = start_time.replace(month=start_time.month - 1)

    end_year = end_time.year
    start_year = start_time.year
    for year in range(start_year, end_year + 1):
        if year == end_year:
            end_month = end_time.month
        else:
            end_month = 12
        if year == start_year:
            start_month = start_time.month - 1
        else:
            start_month = 0
        for month in defaults.months[start_month:end_month]:
            yield str(year), month, None, None


def year_month_day_index_gen(start_time, end_time):

    end_year = end_time.year
    start_year = start_time.year

    for year in range(start_year, end_year + 1):
        if year == end_year:
            end_month = end_time.month
        else:
            end_month = 12
        if year == start_year:
            start_month = start_time.month - 1
        else:
            start_month = 0

        for month in defaults.months[start_month:end_month]:
            for day in range(1, monthrange(int(year), int(month))[1] + 1):
                if ((day < start_time.day and int(month) == start_time.month and year == start_year)
                        or (day > end_time.day and int(month) == end_time.month and year == end_year)):
                    continue
                for hour in range(23, -1, -1):
                    if (hour < start_time.hour and int(month) == start_time.month and year == start_year
                        and start_time.day == day) \
                            or (hour > end_time.hour  and int(month) == end_time.month and year == end_year and
                                end_time.day == day):
                        continue
                    for minute in range(55, -1, -5):
                        index = str(hour).zfill(2) + str(minute).zfill(2)
                        yield str(year), month, str(day).zfill(2), index